// Code generated by mockery v2.14.0. DO NOT EDIT.

package gocb

import (
	gocbcore "github.com/couchbase/gocbcore/v10"
	mock "github.com/stretchr/testify/mock"
)

// mockKvCapabilityVerifier is an autogenerated mock type for the kvCapabilityVerifier type
type mockKvCapabilityVerifier struct {
	mock.Mock
}

// BucketCapabilityStatus provides a mock function with given fields: cap
func (_m *mockKvCapabilityVerifier) BucketCapabilityStatus(cap gocbcore.BucketCapability) gocbcore.BucketCapabilityStatus {
	ret := _m.Called(cap)

	var r0 gocbcore.BucketCapabilityStatus
	if rf, ok := ret.Get(0).(func(gocbcore.BucketCapability) gocbcore.BucketCapabilityStatus); ok {
		r0 = rf(cap)
	} else {
		r0 = ret.Get(0).(gocbcore.BucketCapabilityStatus)
	}

	return r0
}

type mockConstructorTestingTnewMockKvCapabilityVerifier interface {
	mock.TestingT
	Cleanup(func())
}

// newMockKvCapabilityVerifier creates a new instance of mockKvCapabilityVerifier. It also registers a testing interface on the mock and a cleanup function to assert the mocks expectations.
func newMockKvCapabilityVerifier(t mockConstructorTestingTnewMockKvCapabilityVerifier) *mockKvCapabilityVerifier {
	mock := &mockKvCapabilityVerifier{}
	mock.Mock.Test(t)

	t.Cleanup(func() { mock.AssertExpectations(t) })

	return mock
}
